import numpy as np
from sklearn.datasets import fetch_openml

from network import ConvNet
from solver import Solver


def load_fashion_mnist(flatten=False):
    data = {}
    mnist = fetch_openml(name="Fashion-MNIST")
    X = mnist.data
    y = mnist.target.astype(np.uint8)

    #######################################################################
    # Optional: you're free to preprocess images here                     #
    #######################################################################
    pass
    #######################################################################
    #                         END OF YOUR CODE                            #
    #######################################################################

    if not flatten:
        X = X.reshape(X.shape[0], 28, 28)
        X = X[:, np.newaxis, :, :]

    #######################################################################
    # Optional: you're free to adjust the training and val split.         #
    # However, the last 10000 images must be the test set.                #
    #######################################################################
    data['X_train'] = X[:50000]
    data['y_train'] = y[:50000]
    data['X_val'] = X[50000:60000]
    data['y_val'] = y[50000:60000]
    data['X_test'] = X[60000:]
    data['y_test'] = y[60000:]
    return data


def train():
    # load data
    data = load_fashion_mnist()

    # initialize model
    model = ConvNet()

    # intialize solver
    solver = Solver(model, data, update_rule='sgd',
                    optim_config={'learning_rate': 1e-2},
                    lr_decay=1.0, num_epochs=1,
                    batch_size=16, print_every=1)

    # start training
    solver.train()

    # report test accuracy
    acc = solver.check_accuracy(data['X_test'], data['y_test'])
    print("Test accuracy: {}".format(acc))


if __name__=="__main__":
    train()
